@@ -12,6 +12,7 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <unistd.h>
+#include <stdbool.h>
                      
 #include <netlink/genl/genl.h>
 #include <netlink/genl/family.h>
@@ -21,7 +22,6 @@
 
 #include "nl80211.h"
 #include "iw.h"
-#include "version.h"
 
 #ifndef CONFIG_LIBNL20
 /* libnl 2.0 compatibility code */
@@ -31,12 +31,12 @@ static inline struct nl_handle *nl_socket_alloc(void)
 	return nl_handle_alloc();
 }
 
-static inline void nl_socket_free(struct nl_handle *h)
+static inline void nl_socket_free(struct nl_sock *h)
 {
 	nl_handle_destroy(h);
 }
 
-static inline int __genl_ctrl_alloc_cache(struct nl_handle *h, struct nl_cache **cache)
+static inline int __genl_ctrl_alloc_cache(struct nl_sock *h, struct nl_cache **cache)
 {
 	struct nl_cache *tmp = genl_ctrl_alloc_cache(h);
 	if (!tmp)
@@ -47,25 +47,25 @@ static inline int __genl_ctrl_alloc_cache(struct nl_handle *h, struct nl_cache *
 #define genl_ctrl_alloc_cache __genl_ctrl_alloc_cache
 #endif /* CONFIG_LIBNL20 */
 
-static int debug = 0;
+int iw_debug = 0;
 
 static int nl80211_init(struct nl80211_state *state)
 {
 	int err;
 
-	state->nl_handle = nl_socket_alloc();
-	if (!state->nl_handle) {
-		fprintf(stderr, "Failed to allocate netlink handle.\n");
+	state->nl_sock = nl_socket_alloc();
+	if (!state->nl_sock) {
+		fprintf(stderr, "Failed to allocate netlink socket.\n");
 		return -ENOMEM;
 	}
 
-	if (genl_connect(state->nl_handle)) {
+	if (genl_connect(state->nl_sock)) {
 		fprintf(stderr, "Failed to connect to generic netlink.\n");
 		err = -ENOLINK;
 		goto out_handle_destroy;
 	}
 
-	if (genl_ctrl_alloc_cache(state->nl_handle, &state->nl_cache)) {
+	if (genl_ctrl_alloc_cache(state->nl_sock, &state->nl_cache)) {
 		fprintf(stderr, "Failed to allocate generic netlink cache.\n");
 		err = -ENOMEM;
 		goto out_handle_destroy;
@@ -83,7 +83,7 @@ static int nl80211_init(struct nl80211_state *state)
  out_cache_free:
 	nl_cache_free(state->nl_cache);
  out_handle_destroy:
-	nl_socket_free(state->nl_handle);
+	nl_socket_free(state->nl_sock);
 	return err;
 }
 
@@ -91,54 +91,105 @@ static void nl80211_cleanup(struct nl80211_state *state)
 {
 	genl_family_put(state->nl80211);
 	nl_cache_free(state->nl_cache);
-	nl_socket_free(state->nl_handle);
+	nl_socket_free(state->nl_sock);
 }
 
-__COMMAND(NULL, NULL, NULL, 0, 0, 0, CIB_NONE, NULL);
-__COMMAND(NULL, NULL, NULL, 1, 0, 0, CIB_NONE, NULL);
+__COMMAND(NULL, NULL, "", NULL, 0, 0, 0, CIB_NONE, NULL, NULL);
+__COMMAND(NULL, NULL, "", NULL, 1, 0, 0, CIB_NONE, NULL, NULL);
 
 static int cmd_size;
 
-static void usage(const char *argv0)
+static void __usage_cmd(struct cmd *cmd, char *indent, bool full)
+{
+	const char *start, *lend, *end;
+
+	fprintf(stderr, "%s", indent);
+
+	switch (cmd->idby) {
+	case CIB_NONE:
+		break;
+	case CIB_PHY:
+		fprintf(stderr, "phy <phyname> ");
+		break;
+	case CIB_NETDEV:
+		fprintf(stderr, "dev <devname> ");
+		break;
+	}
+	if (cmd->section)
+		fprintf(stderr, "%s ", cmd->section);
+	fprintf(stderr, "%s", cmd->name);
+	if (cmd->args)
+		fprintf(stderr, " %s", cmd->args);
+	fprintf(stderr, "\n");
+
+	if (!full || !cmd->help)
+		return;
+
+	/* hack */
+	if (strlen(indent))
+		indent = "\t\t";
+	else
+		fprintf(stderr, "\n");
+
+	/* print line by line */
+	start = cmd->help;
+	end = strchr(start, '\0');
+	do {
+		lend = strchr(start, '\n');
+		if (!lend)
+			lend = end;
+		fprintf(stderr, "%s", indent);
+		fprintf(stderr, "%.*s\n", (int)(lend - start), start);
+		start = lend + 1;
+	} while (end != lend);
+
+	fprintf(stderr, "\n");
+}
+
+static void usage_options(void)
+{
+	fprintf(stderr, "Options:\n");
+	fprintf(stderr, "\t--debug\t\tenable netlink debugging\n");
+}
+
+static const char *argv0;
+
+static void usage(bool full)
 {
 	struct cmd *cmd;
 
 	fprintf(stderr, "Usage:\t%s [options] command\n", argv0);
-	fprintf(stderr, "Options:\n");
-	fprintf(stderr, "\t--debug\t\tenable netlink debugging\n");
-	fprintf(stderr, "\t--version\tshow version\n");
+	usage_options();
+	fprintf(stderr, "\t--version\tshow version (%s)\n", iw_version);
 	fprintf(stderr, "Commands:\n");
-	fprintf(stderr, "\thelp\n");
-	fprintf(stderr, "\tevent\n");
 	for (cmd = &__start___cmd; cmd < &__stop___cmd;
 	     cmd = (struct cmd *)((char *)cmd + cmd_size)) {
 		if (!cmd->handler || cmd->hidden)
 			continue;
-		switch (cmd->idby) {
-		case CIB_NONE:
-			fprintf(stderr, "\t");
-			/* fall through */
-		case CIB_PHY:
-			if (cmd->idby == CIB_PHY)
-				fprintf(stderr, "\tphy <phyname> ");
-			/* fall through */
-		case CIB_NETDEV:
-			if (cmd->idby == CIB_NETDEV)
-				fprintf(stderr, "\tdev <devname> ");
-			if (cmd->section)
-				fprintf(stderr, "%s ", cmd->section);
-			fprintf(stderr, "%s", cmd->name);
-			if (cmd->args)
-				fprintf(stderr, " %s", cmd->args);
-			fprintf(stderr, "\n");
-			break;
-		}
+		__usage_cmd(cmd, "\t", full);
 	}
 }
 
+static int print_help(struct nl80211_state *state,
+		      struct nl_cb *cb,
+		      struct nl_msg *msg,
+		      int argc, char **argv)
+{
+	exit(3);
+}
+TOPLEVEL(help, NULL, 0, 0, CIB_NONE, print_help,
+	 "Print usage for each command.");
+
+static void usage_cmd(struct cmd *cmd)
+{
+	fprintf(stderr, "Usage:\t%s [options] ", argv0);
+	__usage_cmd(cmd, "", true);
+	usage_options();
+}
+
 static void version(void)
 {
-	printf("iw version " IW_VERSION "\n");
+	printf("iw version %s\n", iw_version);
 }
 
 static int phy_lookup(char *name)
@@ -149,6 +200,8 @@ static int phy_lookup(char *name)
 	snprintf(buf, sizeof(buf), "/sys/class/ieee80211/%s/index", name);
 
 	fd = open(buf, O_RDONLY);
+	if (fd < 0)
+		return -1;
 	pos = read(fd, buf, sizeof(buf) - 1);
 	if (pos < 0)
 		return -1;
@@ -178,28 +231,44 @@ static int ack_handler(struct nl_msg *msg, void *arg)
 	return NL_STOP;
 }
 
-static int handle_cmd(struct nl80211_state *state,
-		      enum command_identify_by idby,
-		      int argc, char **argv)
+static int __handle_cmd(struct nl80211_state *state, enum id_input idby,
+			int argc, char **argv, struct cmd **cmdout)
 {
-	struct cmd *cmd;
-	struct nl_cb *cb = NULL;
+	struct cmd *cmd, *match = NULL;
+	struct nl_cb *cb;
 	struct nl_msg *msg;
 	int devidx = 0;
-	int err;
+	int err, o_argc;
 	const char *command, *section;
+	char *tmp, **o_argv;
+	enum command_identify_by command_idby = CIB_NONE;
 
-	if (argc <= 1 && idby != CIB_NONE)
+	if (argc <= 1 && idby != II_NONE)
 		return 1;
 
+	o_argc = argc;
+	o_argv = argv;
+
 	switch (idby) {
-	case CIB_PHY:
+	case II_PHY_IDX:
+		command_idby = CIB_PHY;
+		devidx = strtoul(*argv + 4, &tmp, 0);
+		if (*tmp != '\0')
+			return 1;
+		argc--;
+		argv++;
+		break;
+	case II_PHY_NAME:
+		command_idby = CIB_PHY;
 		devidx = phy_lookup(*argv);
 		argc--;
 		argv++;
 		break;
-	case CIB_NETDEV:
+	case II_NETDEV:
+		command_idby = CIB_NETDEV;
 		devidx = if_nametoindex(*argv);
+		if (devidx == 0)
+			devidx = -1;
 		argc--;
 		argv++;
 		break;
@@ -207,6 +276,9 @@ static int handle_cmd(struct nl80211_state *state,
 		break;
 	}
 
+	if (devidx < 0)
+		return -errno;
+
 	section = command = *argv;
 	argc--;
 	argv++;
@@ -215,15 +287,18 @@ static int handle_cmd(struct nl80211_state *state,
 	     cmd = (struct cmd *)((char *)cmd + cmd_size)) {
 		if (!cmd->handler)
 			continue;
-		if (cmd->idby != idby)
+		if (cmd->idby != command_idby)
 			continue;
 		if (cmd->section) {
 			if (strcmp(cmd->section, section))
 				continue;
 			/* this is a bit icky ... */
 			if (command == section) {
-				if (argc <= 0)
+				if (argc <= 0) {
+					if (match)
+						break;
 					return 1;
+				}
 				command = *argv;
 				argc--;
 				argv++;
@@ -234,19 +309,31 @@ static int handle_cmd(struct nl80211_state *state,
 			continue;
 		if (argc && !cmd->args)
 			continue;
-		break;
+
+		match = cmd;
 	}
 
-	if (cmd >= &__stop___cmd)
+	cmd = match;
+
+	if (!cmd)
 		return 1;
 
+	if (cmdout)
+		*cmdout = cmd;
+
+	if (!cmd->cmd) {
+		argc = o_argc;
+		argv = o_argv;
+		return cmd->handler(state, NULL, NULL, argc, argv);
+	}
+
 	msg = nlmsg_alloc();
 	if (!msg) {
 		fprintf(stderr, "failed to allocate netlink message\n");
 		return 2;
 	}
 
-	cb = nl_cb_alloc(debug ? NL_CB_DEBUG : NL_CB_DEFAULT);
+	cb = nl_cb_alloc(iw_debug ? NL_CB_DEBUG : NL_CB_DEFAULT);
 	if (!cb) {
 		fprintf(stderr, "failed to allocate netlink callbacks\n");
 		err = 2;
@@ -256,7 +343,7 @@ static int handle_cmd(struct nl80211_state *state,
 	genlmsg_put(msg, 0, 0, genl_family_get_id(state->nl80211), 0,
 		    cmd->nl_msg_flags, cmd->cmd, 0);
 
-	switch (idby) {
+	switch (command_idby) {
 	case CIB_PHY:
 		NLA_PUT_U32(msg, NL80211_ATTR_WIPHY, devidx);
 		break;
@@ -267,11 +354,11 @@ static int handle_cmd(struct nl80211_state *state,
 		break;
 	}
 
-	err = cmd->handler(cb, msg, argc, argv);
+	err = cmd->handler(state, cb, msg, argc, argv);
 	if (err)
 		goto out;
 
-	err = nl_send_auto_complete(state->nl_handle, msg);
+	err = nl_send_auto_complete(state->nl_sock, msg);
 	if (err < 0)
 		goto out;
 
@@ -282,7 +369,7 @@ static int handle_cmd(struct nl80211_state *state,
 	nl_cb_set(cb, NL_CB_ACK, NL_CB_CUSTOM, ack_handler, &err);
 
 	while (err > 0)
-		nl_recvmsgs(state->nl_handle, cb);
+		nl_recvmsgs(state->nl_sock, cb);
  out:
 	nl_cb_put(cb);
  out_free_msg:
@@ -293,79 +380,27 @@ static int handle_cmd(struct nl80211_state *state,
 	return 2;
 }
 
-static int no_seq_check(struct nl_msg *msg, void *arg)
+int handle_cmd(struct nl80211_state *state, enum id_input idby,
+	       int argc, char **argv)
 {
-	return NL_OK;
-}
-
-static int print_event(struct nl_msg *msg, void *arg)
-{
-	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
-	struct nlattr *tb[NL80211_ATTR_MAX + 1];
-
-	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
-		  genlmsg_attrlen(gnlh, 0), NULL);
-                          
-	switch (gnlh->cmd) {
-	case NL80211_CMD_NEW_WIPHY:
-		printf("wiphy rename: phy #%d to %s\n",
-		       nla_get_u32(tb[NL80211_ATTR_WIPHY]),
-		       nla_get_string(tb[NL80211_ATTR_WIPHY_NAME]));
-		break;
-	default:
-		printf("unknown event: %d\n", gnlh->cmd);
-		break;
-	}
-
-	return NL_SKIP;
-}
-
-static int listen_events(struct nl80211_state *state,
-			 int argc, char **argv)
-{
-	int mcid, ret;
-	struct nl_cb *cb = nl_cb_alloc(debug ? NL_CB_DEBUG : NL_CB_DEFAULT);
-
-	if (!cb) {
-		fprintf(stderr, "failed to allocate netlink callbacks\n");
-		return -ENOMEM;
-	}
-
-	mcid = nl_get_multicast_id(state->nl_handle, "nl80211", "config");
-	if (mcid < 0)
-		return mcid;
-
-	ret = nl_socket_add_membership(state->nl_handle, mcid);
-	if (ret)
-		return ret;
-	
-	/* no sequence checking for multicast messages */
-	nl_cb_set(cb, NL_CB_SEQ_CHECK, NL_CB_CUSTOM, no_seq_check, NULL);
-	nl_cb_set(cb, NL_CB_VALID, NL_CB_CUSTOM, print_event, NULL);
-
-	while (1)
-		nl_recvmsgs(state->nl_handle, cb);
-
-	nl_cb_put(cb);
-
-	return 0;
+	return __handle_cmd(state, idby, argc, argv, NULL);
 }
 
 int main(int argc, char **argv)
 {
 	struct nl80211_state nlstate;
 	int err;
-	const char *argv0;
+	struct cmd *cmd = NULL;
 
 	/* calculate command size including padding */
-	cmd_size = abs((long)&__cmd_NULL_1_CIB_NONE_0
-	             - (long)&__cmd_NULL_0_CIB_NONE_0);
+	cmd_size = abs((long)&__cmd_NULL_NULL_1_CIB_NONE_0
+	             - (long)&__cmd_NULL_NULL_0_CIB_NONE_0);
 	/* strip off self */
 	argc--;
 	argv0 = *argv++;
 
 	if (argc > 0 && strcmp(*argv, "--debug") == 0) {
-		debug = 1;
+		iw_debug = 1;
 		argc--;
 		argv++;
 	}
@@ -375,8 +410,9 @@ int main(int argc, char **argv)
 		return 0;
 	}
 
+	/* need to treat "help" command specially so it works w/o nl80211 */
 	if (argc == 0 || strcmp(*argv, "help") == 0) {
-		usage(argv0);
+		usage(argc != 0);
 		return 0;
 	}
 
@@ -384,22 +420,28 @@ int main(int argc, char **argv)
 	if (err)
 		return 1;
 
-	if (strcmp(*argv, "event") == 0) {
-		err = listen_events(&nlstate, argc, argv);
-	} else if (strcmp(*argv, "dev") == 0) {
+	if (strcmp(*argv, "dev") == 0 && argc > 1) {
 		argc--;
 		argv++;
-		err = handle_cmd(&nlstate, CIB_NETDEV, argc, argv);
-	} else if (strcmp(*argv, "phy") == 0) {
-		argc--;
-		argv++;
-		err = handle_cmd(&nlstate, CIB_PHY, argc, argv);
+		err = __handle_cmd(&nlstate, II_NETDEV, argc, argv, &cmd);
+	} else if (strncmp(*argv, "phy", 3) == 0 && argc > 1) {
+		if (strlen(*argv) == 3) {
+			argc--;
+			argv++;
+			err = __handle_cmd(&nlstate, II_PHY_NAME, argc, argv, &cmd);
+		} else if (*(*argv + 3) == '#')
+			err = __handle_cmd(&nlstate, II_PHY_IDX, argc, argv, &cmd);
+		else
+			err = 1;
 	} else
-		err = handle_cmd(&nlstate, CIB_NONE, argc, argv);
+		err = __handle_cmd(&nlstate, II_NONE, argc, argv, &cmd);
 
-	if (err == 1)
-		usage(argv0);
-	if (err < 0)
+	if (err == 1) {
+		if (cmd)
+			usage_cmd(cmd);
+		else
+			usage(false);
+	} else if (err < 0)
 		fprintf(stderr, "command failed: %s (%d)\n", strerror(-err), err);
 
 	nl80211_cleanup(&nlstate);
